# -*- coding: utf-8 -*-
#                            __
#  _ __  _   _  ___  ___  __| | ___   ___
# | '_ \| | | |/ _ \/ __|/ _` |/ _ \ / __|
# | |_) | |_| |  __/\__ \ (_| | (_) | (__
# | .__/ \__, |\___||___/\__,_|\___/ \___|
# |_|    |___/
#
"""
.. module:: pyesdoc.__init__.py

   :license: GPL / CeCILL
   :platform: Unix, Windows
   :synopsis: Package initializer.

.. moduleauthor:: Earth System Documentation (ES-DOC) <dev@es-doc.org>

"""
__title__ = 'pyesdoc'
__version__ = '0.14.2.0'
__author__ = 'ES-DOC'
__license__ = 'GPL'
__copyright__ = 'Copyright 2018 ES-DOC'


from pyesdoc import archive
from pyesdoc import constants
from pyesdoc import drq
from pyesdoc import exceptions
from pyesdoc import mp
from pyesdoc import ontologies
from pyesdoc import utils

from pyesdoc._associating import associate
from pyesdoc._associating import associate_by_id
from pyesdoc._associating import associate_by_id_and_version
from pyesdoc._associating import associate_by_name

from pyesdoc._extensions import extend
from pyesdoc._extensions import is_extendable

from pyesdoc._factory import create

from pyesdoc._io import convert as convert_file
from pyesdoc._io import get_filename
from pyesdoc._io import read
from pyesdoc._io import seek
from pyesdoc._io import write

from pyesdoc.mp.specializations.utils_cache import get_topic_specialization
from pyesdoc.mp.specializations.utils_cache import get_property_specialization

from pyesdoc._publishing import publish
from pyesdoc._publishing import retrieve
from pyesdoc._publishing import unpublish

from pyesdoc._search import search
from pyesdoc._search import SearchResult
from pyesdoc._search import SearchResultItem

from pyesdoc._security import authenticate_user
from pyesdoc._security import AuthenticationError
from pyesdoc._security import authorize_user
from pyesdoc._security import AuthorizationError
from pyesdoc._security import strip_credentials

from pyesdoc._serialization import convert
from pyesdoc._serialization import decode
from pyesdoc._serialization import encode

from pyesdoc.utils import config
from pyesdoc.utils.logger import log
from pyesdoc.utils.logger import log_error
from pyesdoc.utils.logger import log_warning

from pyesdoc._validation import is_valid
from pyesdoc._validation import validate



"""A realm process sepecialization.

For further information goto http://wordpress.es-doc.org/cmip6-model-specializations.

"""
# --------------------------------------------------------------------
# INTERNAL (do not change)
# --------------------------------------------------------------------
from collections import OrderedDict

DETAILS = OrderedDict()
ENUMERATIONS = OrderedDict()

# --------------------------------------------------------------------
# DESCRIPTION: Short description of the specialization.
# --------------------------------------------------------------------
DESCRIPTION = 'Land surface lakes'

# --------------------------------------------------------------------
# LAKES: top level
# --------------------------------------------------------------------
DETAILS['toplevel'] = {
    'description': 'Land surface lakes top level properties',
    'properties' : [
        ('coupling_with_rivers', 'bool', '1.1',
             'Are lakes coupled to the river routing model component?'),
        ('time_step', 'int', '1.1',
             'Time step of lake scheme in seconds'),
        ('quantities_exchanged_with_rivers', 'ENUM:quantities_exchanged_with_rivers_types', '0.N',
             'If coupling with rivers, which quantities are exchanged between the lakes and rivers'),
        ('vertical_grid', 'str', '0.1',
             'Describe the vertical grid of lakes'),
        ('prognostic_variables', 'cs-str', '1.1',
             'List the prognostic variables of the lake scheme'),
    ]
}

# --------------------------------------------------------------------
# LAKES: process
# --------------------------------------------------------------------
DETAILS['method'] = {
    'description': 'Lakes treatment',
    'properties' : [
        ('ice_treatment', 'bool', '1.1',
             'Is lake ice included?'),
        ('albedo', 'ENUM:lake_albedo_types', '1.1',
             'Describe the treatment of lake albedo'),
        ('dynamics', 'ENUM:lake_dynamics_types', '1.N',
             'Which dynamics of lakes are treated? horizontal, vertical, etc.'),
        ('dynamic_lake_extent', 'bool', '1.1',
             'Is a dynamic lake extent scheme included?'),
        ('endorheic_basins', 'bool', '1.1',
             'Basins not flowing to ocean included?'),
    ]
}

# --------------------------------------------------------------------
# LAKES: Wetlands
# --------------------------------------------------------------------
DETAILS['wetlands'] = {
    'description': 'Welands treatment',
    'properties' : [
        ('description', 'l-str', '0.1',
             'Describe the treatment of wetlands, if any'),
    ]
}

# --------------------------------------------------------------------
# LAKES: ENUMERATIONS
# --------------------------------------------------------------------
ENUMERATIONS['quantities_exchanged_with_rivers_types'] = {
    'description': 'Quantities that are exchanged between the lakes and rivers',
    'is_open': True,
    'members': [
        ('heat', None),
        ('water', None),
        ('tracers', None),
    ]
}

ENUMERATIONS['lake_albedo_types'] = {
    'description': 'Describe the treatment of lake albedo',
    'is_open': True,
    'members': [
        ('prognostic', None),
        ('diagnostic', None),
    ]
}

ENUMERATIONS['lake_dynamics_types'] = {
    'description': 'Which dynamics of lakes are treated?',
    'is_open': True,
    'members': [
        ('No lake dynamics', None),
        ('vertical', None),
        ('horizontal', None),
    ]
}




"""
.. module:: designing_numerical_experiment.py
   :platform: Unix, Windows
   :synopsis: Extends a cim.v1.designing.numerical_experiment document.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""
from pyesdoc import constants



def get_extenders():
    """Returns set of extension functions.

    """
    return (
        _set_summary_fields,
        )


def _set_summary_fields(ctx):
    """Sets related experiment viewer urls.

    """
    ctx.ext.summary_fields = (
        ctx.doc.canonical_id.upper(),
        ctx.doc.long_name
    )



"""A realm grid sepecialization.

For further information goto http://wordpress.es-doc.org/cmip6-model-specializations.

"""
# --------------------------------------------------------------------
# INTERNAL (do not change)
# --------------------------------------------------------------------
from collections import OrderedDict

DETAILS = OrderedDict()
ENUMERATIONS = OrderedDict()

# --------------------------------------------------------------------
# DESCRIPTION: Short description of the specialization.
# --------------------------------------------------------------------
DESCRIPTION = 'Land surface grid'

# --------------------------------------------------------------------
# DISCRETISATION: Description of the numerics of the discretisation.
# --------------------------------------------------------------------
DETAILS['horizontal'] = {
    'description': 'The horizontal grid in the land surface',
    'properties': [
        ('description', 'l-str', '1.1',
             'Describe the general structure of the horizontal grid (not including any tiling)'),
        ('matches_atmosphere_grid', 'bool', '1.1',
             'Does the horizontal grid match the atmosphere?'),
        ]
    }

DETAILS['vertical'] = {
    'description': 'The vertical grid in the soil',
    'properties': [
        ('description', 'l-str', '1.1',
             'Describe the general structure of the  vertical grid in the soil (not including any tiling)'),       
        ('total_depth', 'int', '1.1',
             'The total depth of the soil (in metres)'),
    ]
}

# --------------------------------------------------------------------
# ENUMERATIONS
# --------------------------------------------------------------------



"""
.. module:: utils_cache.py
   :license: GPL/CeCIL
   :platform: Unix, Windows
   :synopsis: Encapsulates access to specialization cache.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""
import os

from utils_loader import get_modules



# Map of specializations by id.
_CACHE = dict()


def set_specialization(spec):
    """Place specialization in cache.

    """
    _CACHE[spec.id.lower()] = spec


def get_topic_specialization(mip_era, topic):
	"""Returns topic specializations.

	:param str mip_era: MIP era that model documentation is related to.
	:param str topic: Specialization topic.

	:result: Topic specializations.
	:rtype: TopicSpecialization

	"""
	# JIT load.
	_set_topic_specialization(mip_era, topic)

	return _CACHE[_get_topic_cache_key(mip_era, topic)]


def get_property_specialization(specialization_id):
	"""Returns property specializations.

	:param str specialization_id: Specialziation identifier.

	:result: Property specialization.
	:rtype: PropertySpecialization

	"""
	if len(specialization_id) == 0:
		return

	# JIT load.
	parts = specialization_id.split('.')
	_set_topic_specialization(parts[0], parts[1])

	return _CACHE[specialization_id.lower()]


def _set_topic_specialization(mip_era, topic):
	"""Loads topic specializations into memory.

	"""
	if _get_topic_cache_key(mip_era, topic) in _CACHE:
		return

	# JIT to avoid circular dependencies.
	from utils_factory import create_specializations

	mip_era = mip_era.strip().lower()
	topic = topic.strip().lower()
	dpath = os.path.join(os.path.dirname(__file__), mip_era)
	modules = get_modules(dpath, topic)
	create_specializations(modules)


def _get_topic_cache_key(mip_era, topic):
	"""Returns topic specialization cache key.

	"""
	mip_era = mip_era.strip().lower()
	topic = topic.strip().lower()

	return "{}.{}".format(mip_era, topic)



"""
.. module:: utils.py
   :license: GPL/CeCIL
   :platform: Unix, Windows
   :synopsis: CMIP6 specialization validation utility functions.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""
def log(msg):
    """Outputs a message to log.

    :param str msg: Logging message.

    """
    if msg.startswith('-'):
        print(msg)
    else:
        print("ES-DOC :: {}".format(msg))


def get_label(name):
    """Returns a name formatted as a label for UI purposes.

    """
    name = name.replace("_", " ")

    return " ".join("{}{}".format(n[0].upper(), n[1:]) for n in name.split(" "))




"""A model key-properties sepecialization.

"""
# --------------------------------------------------------------------
# INTERNAL (do not change)
# --------------------------------------------------------------------
from collections import OrderedDict

DETAILS = OrderedDict()
ENUMERATIONS = OrderedDict()

# --------------------------------------------------------------------
# CONTACT: Set to top-level specialization co-ordinator.
# --------------------------------------------------------------------
CONTACT = 'Guillaume Levavasseur (IPSL)'

# --------------------------------------------------------------------
# AUTHORS: Set to top-level specialization authors (comma delimited).
# --------------------------------------------------------------------
AUTHORS = 'Guillaume Levavasseur (IPSL), Grigory Nikulin (SMHI)'

# --------------------------------------------------------------------
# DESCRIPTION: Short description of the specialization.
# --------------------------------------------------------------------
DESCRIPTION = 'Key properties of the regional model'

# --------------------------------------------------------------------
# KEY PROPERTIES: top level
# --------------------------------------------------------------------
DETAILS['toplevel'] = {
    'description': 'Top level properties of the regional model'
    }

DETAILS['toplevel:model_type'] = {
    'description': "Type of Regional Climate Model",
    'properties': [
        ('model_family', 'ENUM:model_family_type', '1.1',
            "Type of Regional Climate Model"),
        ('details', 'l-str', '0.1',
            "Provide additional details on type of the Regional Climate Model"),
        ]
    }  

DETAILS['toplevel:genealogy'] = {
    'description': 'Genealogy and history of the model',
    'properties':[
        ('year_released', 'str', '1.1',
            'Year the model was released'),
        ('previous_name', 'str', '0.1',
            'Previously known as'),
        ('difference_to_previous', 'l-str', '0.1',
            'Briefly summarize the differences between this model and its previous version, if applicable'),
        ]
    }
    
DETAILS['toplevel:cordex_domain'] = {
    'description':'CORDEX domain',
    'properties': [
        ('domain','ENUM:cordex_domain', '0.1',
            "CORDEX domain"),
        ]
    }         
    
DETAILS['toplevel:model_calendar'] = {
    'description': 'Model calendar',
    'properties':[
        ('calendar','ENUM:calendar_used', '0.1',
            "What calender is uded by the model"),
        ]
    }
    
DETAILS['toplevel:spinup'] = {
    'description':'Spin-up',
    'properties': [
        ('details', 'l-str', '0.1',
            'Describe details if any spin-up is used'),
        ]
    }
    
DETAILS['toplevel:nudging'] = {
    'description':'Nudging',
    'properties': [
        ('details', 'l-str', '0.1',
            'Describe details if spectral nudging is used (variables, levels, etc.)'),
        ]
    }           

DETAILS['toplevel:software_properties'] = {
    'description': 'Software properties of the model',
    'properties':[
        ('repository','str', '0.1',
            "Location of code for this component."),
        ('code_version','str', '0.1',
            "Code version identifier."),
        ('code_languages','cs-str', '0.1',
            "Code language(s)."),
        ('components_structure','str', '0.1',
            "Describe how model realms are structured into independent software components (coupled via a coupler) and internal software components."),
        ('coupler','ENUM:coupler_framework', '0.1',
            "Overarching coupling framework for the model."),
        ]
    }
   
DETAILS['toplevel:flux_correction'] = {
    'description': 'Flux correction properties of the model',
    'properties':[
        ('details', 'l-str', '1.1',
            'Describe if/how flux corrections are applied in the model'),
        ]
    }
    
DETAILS['toplevel:reference'] = {
    'description': 'Latest reference',
    'properties':[
        ('details', 'l-str', '1.1',
            'Provide the latest reference describing the model'),
        ]
    }             

# --------------------------------------------------------------------
# REALM COUPLING: general questions
# --------------------------------------------------------------------
DETAILS['coupling'] = {
    'description': '',
    'properties':[
        ('atmosphere_double_flux', 'bool', '1.1',
             'Is the atmosphere passing a double flux to the ocean and sea ice (as opposed to a single one)?'),
        ('atmosphere_fluxes_calculation_grid', 'ENUM:atmosphere_fluxes_calculation_grid', '0.1',
             'Where are the air-sea fluxes calculated'),
        ('atmosphere_relative_winds', 'bool', '1.1',
            'Are relative or absolute winds used to compute the flux? I.e. do ocean surface currents enter the wind stress calculation?'),
    ]
}

# --------------------------------------------------------------------
# TUNING APPLIED: Any tuning used to optimise the parameters in this realm
# --------------------------------------------------------------------
DETAILS['tuning_applied'] = {
    'description': 'Tuning methodology for model',
    'properties': [
        ('description', 'l-str', '1.1',
            "General overview description of tuning: explain and motivate the main targets and metrics/diagnostics retained. Document the relative weight given to climate performance metrics/diagnostics versus process oriented metrics/diagnostics, and on the possible conflicts with parameterization level tuning. In particular describe any struggle with a parameter value that required pushing it to its limits to solve a particular model deficiency."),
        ('regional_metrics_used', 'cs-str', '0.1',
            "List of regional metrics/diagnostics of mean state (e.g THC, AABW, regional means etc) used in tuning model/component"),
        ('trend_metrics_used', 'cs-str', '0.1',
            "List observed trend metrics/diagnostics used in tuning model/component (such as 20th century)"),
        ('energy_balance','str', '1.1',
            "Describe how energy balance was obtained in the full system: in the various components independently or at the components coupling stage?"),
        ('fresh_water_balance','str', '1.1',
            "Describe how fresh_water balance was obtained in the full system: in the various components independently or at the components coupling stage?")
        ]
    }

# --------------------------------------------------------------------
# GCM FORCING: GCM forcing used to drive RCM
# --------------------------------------------------------------------
DETAILS['gcm_forcing'] = {
    'description': 'GCM forcing'
}

DETAILS['gcm_forcing:atmos_lbc_eval'] = {
    'description': 'Atmospheric Lateral Boundary Conditions for the evaluation experiment',
    'properties': [
        ('atmos_lbc_eval_variables', 'l-str', '0.1',
            'What atmospheric variables are used to drive the regional model'),
        ('atmos_lbc_eval_frequency', 'l-str', '0.1',
            'At what frequency'),
        ('atmos_lbc_eval_top', 'l-str', '0.1',
            'Top boundary conditions'),
        ('additional_information', 'l-str', '0.1',
            'Additional information relating to the implementation of the Atmospheric Lateral Boundary Conditions for the evaluation experiment (e.g. citations, use of non-standard datasets, etc.).'),
    ]
}

DETAILS['gcm_forcing:atmos_lbc_scn'] = {
    'description': 'Atmospheric Lateral Boundary Conditions for the historical and scenario experiments',
    'properties': [
        ('atmos_lbc_scn_variables', 'l-str', '0.1',
            'What atmospheric variables are used to drive the regional model'),
        ('atmos_lbc_scn_frequency', 'l-str', '0.1',
            'At what frequency'),
        ('atmos_lbc_scn_top', 'l-str', '0.1',
            'Top boundary conditions'),
        ('additional_information', 'l-str', '0.1',
            'Additional information relating to the implementation of the Atmospheric Lateral Boundary Conditions for the historical and scenario experiments (e.g. citations, use of non-standard datasets, etc.).'),
    ]
}

DETAILS['gcm_forcing:ocean_lbc_eval'] = {
    'description': 'Lower Boundary Conditions for the evaluation experiment',
    'properties': [
        ('sst_lbc_eval_variable', 'l-str', '0.1',
            'What sea surface temperature is used to drive the regional model (driving reanalysis or observations)'),
        ('sst_lbc_eval_frequency', 'l-str', '0.1',
            'At what frequency'),
        ('sic_lbc_eval_variable', 'l-str', '0.1',
            'What sea-ice concentration is used to drive the regional model (driving reanalysis or observations)'),
        ('sic_lbc_eval_frequency', 'l-str', '0.1',
            'At what frequency'),
        ('additional_information', 'l-str', '0.1',
            'Additional information relating to the implementation of the Lower Boundary Conditions for the evaluation experiment (e.g. citations, use of non-standard datasets, etc.)'),
    ]
}

DETAILS['gcm_forcing:ocean_lbc_scn'] = {
    'description': 'Lower Boundary Conditions for the historical and scenario experiment',
    'properties': [
        ('sst_lbc_scn_frequency', 'l-str', '0.1',
            'At what frequency is sea surface temperature taken'),
        ('sic_lbc_scn_frequency', 'l-str', '0.1',
            'At what frequency is sea-ice concentration taken'),
        ('additional_information', 'l-str', '0.1',
            'Additional information relating to the implementation of the Lower Lateral Boundary Conditions for the historical and scenario experiments (e.g. citations, use of non-standard datasets, etc.).'),
    ]
}

DETAILS['gcm_forcing:aerosol_eval'] = {
    'description': 'Aerosol datasets and parameters used for the evaluation experiment',
    'properties': [
        ('aerosol_eval_dataset', 'l-str', '0.1',
            'What aerosol datasets and parameters are used to drive the regional model (driving reanalysis or observations)'),
        ('aerosol_eval_frequency', 'l-str', '0.1',
            'At what frequency (climatology, transient, etc.)'),
        ('additional_information', 'l-str', '0.1',
            'Additional information relating to the implementation of aerosol for the evaluation experiment'),
    ]
}

DETAILS['gcm_forcing:aerosol_scn'] = {
    'description': 'Aerosol datasets and parameters are used for the historical and scenario experiments',
    'properties': [
        ('aerosol_scn_dataset', 'l-str', '0.1',
            'What aerosol datasets and parameters are used to drive the regional model'),
        ('aerosol_scn_frequency', 'l-str', '0.1',
            'At what frequency (climatology, transient, etc.)'),
        ('additional_information', 'l-str', '0.1',
            'Additional information relating to the implementation of aerosol for the historical and scenario experiment'),
    ]
}

DETAILS['gcm_forcing:ozone_eval'] = {
    'description': 'Ozone datasets used for the evaluation experiment',
    'properties': [
        ('ozone_eval_dataset', 'l-str', '0.1',
            'What ozone dataset is used to drive the regional model (driving reanalysis or observations)'),
        ('ozone_eval_frequency', 'l-str', '0.1',
            'At what frequency (climatology, transient, etc.)'),
        ('additional_information', 'l-str', '0.1',
            'Additional information relating to the implementation of ozone for the evaluation experiment'),
    ]
}

DETAILS['gcm_forcing:ozone_scn'] = {
    'description': 'Ozone datasets used for the historical and scenario experiments',
    'properties': [
        ('ozone_scn_dataset', 'l-str', '0.1',
            'What ozone dataset is used to drive the regional model'),
        ('ozone_scn_frequency', 'l-str', '0.1',
            'At what frequency (climatology, transient, etc.)'),
        ('additional_information', 'l-str', '0.1',
            'Additional information relating to the implementation of ozone for the historical and scenario experiments'),
    ]
}

DETAILS['gcm_forcing:land_cover_eval'] = {
    'description': 'Land cover datasets used for the evaluation experiment',
    'properties': [
        ('land_cover_eval_dataset', 'l-str', '0.1',
            'What land cover dataset is used in the regional model'),
        ('land_cover_eval_frequency', 'l-str', '0.1',
            'At what frequency (climatology, transient, etc.)'),
        ('additional_information', 'l-str', '0.1',
            'Additional information relating to the implementation of land cover for the evaluation experiment'),
    ]
}

DETAILS['gcm_forcing:land_cover_scn'] = {
    'description': 'Land cover datasets used for the historical and scenario experiments',
    'properties': [
        ('land_cover_scn_dataset', 'l-str', '0.1',
            'What land cover dataset is used in the regional model'),
        ('land_cover_scn_frequency', 'l-str', '0.1',
            'At what frequency (climatology, transient, etc.)'),
        ('additional_information', 'l-str', '0.1',
            'Additional information relating to the implementation of land cover for tthe historical and scenario experiments'),
    ]
}

# --------------------------------------------------------------------
# KEY PROPERTIES: ENUMERATIONS
# --------------------------------------------------------------------
ENUMERATIONS['model_family_type'] = {
    'description': 'Type of regional model',
    'is_open': True,
    'members': [
        ('Regional Atmospheric Climate Model', None),
        ('Regional Ocean Atmosphere Coupled Climate Model', None),
        ('Other', None),
        ]
    }
    
ENUMERATIONS['cordex_domain'] = {
    'description': 'Select the CORDEX domain',
    'is_open': True,
    'members': [
        ("EUR-11", None),
    ]
}

ENUMERATIONS['calendar_used'] = {
    'description': 'What calendar is used by the model',
    'is_open': True,
    'members': [
        ("GCM-based", None),
        ("Standard", None),
        
    ]
}    

ENUMERATIONS['coupler_framework'] = {
    'description': 'Coupling software framework.',
    'is_open': True,
    'members': [
        ("OASIS", "The OASIS coupler - prior to OASIS-MCT"),
        ("OASIS3-MCT", "The MCT variant of the OASIS coupler"),
        ("ESMF", "Vanilla Earth System Modelling Framework"),
        ("NUOPC", "National Unified Operational Prediction Capability variant of ESMF"),
        ("Bespoke", "Customised coupler developed for this model"),
        ("Unknown", "It is not known what/if-a coupler is used"),
        ("None", "No coupler is used")
        ]
    }

ENUMERATIONS['atmosphere_fluxes_calculation_grid'] = {
    'description': 'Where are the air-sea fluxes calculated',
    'is_open': True,
    'members': [
        ("Atmosphere grid", None),
        ("Ocean grid", None),
        ("Specific coupler grid", None),
    ]
}



"""A process specialization.

For further information goto http://wordpress.es-doc.org/cmip6-model-specializations.

"""
# --------------------------------------------------------------------
# INTERNAL (do not change)
# --------------------------------------------------------------------
from collections import OrderedDict

DETAILS = OrderedDict()
ENUMERATIONS = OrderedDict()

# --------------------------------------------------------------------
# DESCRIPTION: Short description of the specialization.
# --------------------------------------------------------------------
DESCRIPTION = 'Atmospheric chemistry photo chemistry'

# --------------------------------------------------------------------
# PROCESS: top level
# --------------------------------------------------------------------
DETAILS['toplevel'] = {
    'description': '',
    'properties': [
        ('number_of_reactions', 'int', '1.1',
             'The number of reactions in the photo-chemistry scheme.'),
    ],
}

# --------------------------------------------------------------------
# SUB-PROCESS: Photolysis
# --------------------------------------------------------------------
DETAILS['photolysis'] = {
    'description': 'Photolysis scheme',
    'properties': [
        ('method', 'ENUM:photolysis_methods', '1.1',
            'Photolysis scheme'),
        ('environmental_conditions', 'str', '0.1',
             'Describe any environmental conditions taken into account by the photolysis scheme (e.g. whether pressure- and temperature-sensitive cross-sections and quantum yields in the photolysis calculations are modified to reflect the modelled conditions.)'),
    ],
}

# --------------------------------------------------------------------
# ENUMERATIONS
# --------------------------------------------------------------------
ENUMERATIONS['photolysis_methods'] = {
    'description': 'Photolysis scheme',
    'is_open': False,
    'members': [
        ('Offline (clear sky)', None),
        ('Offline (with clouds)', None),
        ('Online', None),
    ]
}

ENUMERATIONS['online_processes'] = {
    'description': 'Processes included in the photolysis scheme.',
    'is_open': False,
    'members': [
        ('Scattering', None),
        ('Radiative transfer', None),
        ('Spectral resolution', None),
        ('Impact of clouds', None),
    ]
}



"""A process specialization.

For further information goto http://wordpress.es-doc.org/cmip6-model-specializations.

"""
# --------------------------------------------------------------------
# INTERNAL (do not change)
# --------------------------------------------------------------------
from collections import OrderedDict

DETAILS = OrderedDict()
ENUMERATIONS = OrderedDict()

# --------------------------------------------------------------------
# DESCRIPTION: Short description of the specialization.
# --------------------------------------------------------------------
DESCRIPTION = 'Aerosol optical and radiative properties'

# --------------------------------------------------------------------
# SUB-PROCESS: Absorption
# --------------------------------------------------------------------
DETAILS['absorption'] = {
    'description': 'Absortion properties in aerosol scheme',
    'properties': [
        ('black_carbon', 'float', '0.1',
             'Absorption mass coefficient of black carbon at 550nm (if non-absorbing enter 0)'),
        ('dust', 'float', '0.1',
             'Absorption mass coefficient of dust at 550nm (if non-absorbing enter 0)'),
        ('organics', 'float', '0.1',
             'Absorption mass coefficient of organics at 550nm (if non-absorbing enter 0)'),
    ],
}

# --------------------------------------------------------------------
# SUB-PROCESS: Mixtures
# --------------------------------------------------------------------
DETAILS['mixtures'] = {
    'description': '',
    'properties': [
        ('external', 'bool', '1.1',
            'Is there external mixing with respect to chemical composition?'),
        ('internal', 'bool', '1.1',
            'Is there internal mixing with respect to chemical composition?'),
        ('mixing_rule', 'str', '0.1',
             'If there is internal mixing with respect to chemical composition then indicate the mixing rule'),
    ],
}

# --------------------------------------------------------------------
# SUB-PROCESS: Impact of H2O
# --------------------------------------------------------------------
DETAILS['impact_of_h2o'] = {
    'description': 'The impact of H2O on aerosols',
    'properties': [
        ('size', 'bool', '1.1',
            'Does H2O impact size?'),
        ('internal_mixture', 'bool', '1.1',
            'Does H2O impact aerosol internal mixture?'),
        ('external_mixture', 'bool', '1.1',
            'Does H2O impact aerosol external mixture?'),
    ],
}

# --------------------------------------------------------------------
# SUB-PROCESS: Radiative scheme for aerosol
# --------------------------------------------------------------------
DETAILS['radiative_scheme'] = {
    'description': 'Radiative scheme for aerosol',
    'properties': [
        ('overview', 'l-str', '1.1',
             'Overview of radiative scheme'),
        ('shortwave_bands', 'int', '1.1',
            'Number of shortwave bands'),
        ('longwave_bands', 'int', '1.1',
            'Number of longwave bands'),
    ],
}

# --------------------------------------------------------------------
# SUB-PROCESS: Radiative scheme for aerosol
# --------------------------------------------------------------------
DETAILS['cloud_interactions'] = {
    'description': 'Aerosol-cloud interactions',
    'properties': [
        ('overview', 'l-str', '1.1',
             'Overview of aerosol-cloud interactions'),
        ('twomey', 'bool', '1.1',
            'Is the Twomey effect included?'),
        ('twomey_minimum_ccn', 'int', '0.1',
            'If the Twomey effect is included, then what is the minimum CCN number?'),
        ('drizzle', 'bool', '1.1',
            'Does the scheme affect drizzle?'),
        ('cloud_lifetime', 'bool', '1.1',
            'Does the scheme affect cloud lifetime?'),
        ('longwave_bands', 'int', '1.1',
            'Number of longwave bands'),
    ],
}

# --------------------------------------------------------------------
# ENUMERATIONS
# --------------------------------------------------------------------



"""A realm key-properties specialization.

For further information goto http://wordpress.es-doc.org/cmip6-model-specializations.

"""
# --------------------------------------------------------------------
# INTERNAL (do not change)
# --------------------------------------------------------------------
from collections import OrderedDict

DETAILS = OrderedDict()
ENUMERATIONS = OrderedDict()

# --------------------------------------------------------------------
# DESCRIPTION: Short description of the specialization.
# --------------------------------------------------------------------
DESCRIPTION = 'Ocean Biogeochemistry key properties'

# --------------------------------------------------------------------
# KEY PROPERTIES: top level
# --------------------------------------------------------------------
DETAILS['toplevel'] = {
    'description': 'General key properties in ocean biogeochemistry',
    'properties': [
        ('model_type', 'ENUM:model_family', '1.1',
            'Type of ocean biogeochemistry model'),
        ('elemental_stoichiometry', 'ENUM:elemental_stoichiometry_types', '1.1',
            'Describe elemental stoichiometry (fixed, variable, mix of the two)',),
        ('elemental_stoichiometry_details', 'cs-str', '1.1',
            'Describe which elements have fixed/variable stoichiometry',),
        ('prognostic_variables', 'cs-str', '1.1',
            'List of all prognostic tracer variables in the ocean biogeochemistry component'),
        ('diagnostic_variables', 'cs-str', '1.1',
            'List of all diagnotic tracer variables in the ocean biogeochemistry component (derived from prognostic variables'),
        ('damping', 'str', '0.1',
            'Describe any tracer damping used (such as artificial correction or relaxation to climatology,...)'),
        ]
    }

# --------------------------------------------------------------------
# KEY PROPERTIES: details
# --------------------------------------------------------------------
DETAILS['time_stepping_framework'] = {
    'description': 'Time stepping framework for ocean biogeochemistry',
    'properties' : []
    }

DETAILS['time_stepping_framework:passive_tracers_transport'] = {
    'description': 'Time stepping method for passive tracers transport in ocean biogeochemistry',
    'properties' : [
        ('method', 'ENUM:passive_tracers_transport', '1.1',
            'Time stepping framework for passive tracers'),
        ('timestep_if_not_from_ocean', 'int', '0.1',
            'Time step for passive tracers (if different from ocean)'),
        ]
    }

DETAILS['time_stepping_framework:biology_sources_sinks'] = {
    'description': 'Time stepping framework for biology sources and sinks in ocean biogeochemistry',
    'properties' : [
        ('method', 'ENUM:passive_tracers_transport', '1.1',
            'Time stepping framework for biology sources and sinks'),
        ('timestep_if_not_from_ocean', 'int', '0.1',
            'Time step for biology sources and sinks (if different from ocean)'),
        ]
    }

DETAILS['transport_scheme'] = {
    'description': 'Transport scheme in ocean biogeochemistry',
    'properties' : [
        ('type', 'ENUM:transport_types', '1.1',
            'Type of transport scheme'),
        ('scheme', 'ENUM:transport_scheme', '1.1',
            'Transport scheme used'),
        ('use_different_scheme', 'l-str', '0.1',
            'Decribe transport scheme if different than that of ocean model'),
        ]
    }

DETAILS['boundary_forcing'] = {
    'description': 'Properties of biogeochemistry boundary forcing',
    'properties': [
        ('atmospheric_deposition', 'ENUM:sources_atmos_deposition', '1.1',
            'Describe how atmospheric deposition is modeled'),
        ('river_input', 'ENUM:sources_river_input', '1.1',
            'Describe how river input is modeled'),
        ('sediments_from_boundary_conditions', 'cs-str', '0.1',
            'List which sediments are speficied from boundary condition'),
        ('sediments_from_explicit_model', 'cs-str', '0.1',
            'List which sediments are speficied from explicit sediment model'),
        ]
    }

DETAILS['gas_exchange'] = {
    'description': 'Properties of gas exchange in ocean biogeochemistry ',
    'properties': [
        ('CO2_exchange_present', 'bool', '1.1',
            'Is CO2 gas exchange modeled ?'),
        ('CO2_exchange_type', 'ENUM:gas_exchange_types', '0.1',
            'Describe CO2 gas exchange'),
        ('O2_exchange_present', 'bool', '1.1',
            'Is O2 gas exchange modeled ?'),
        ('O2_exchange_type', 'ENUM:gas_exchange_types', '0.1',
            'Describe O2 gas exchange'),
        ('DMS_exchange_present', 'bool', '1.1',
            'Is DMS gas exchange modeled ?'),
        ('DMS_exchange_type', 'str', '0.1',
            'Specify DMS gas exchange scheme type'),
        ('N2_exchange_present', 'bool', '1.1',
            'Is N2 gas exchange modeled ?'),
        ('N2_exchange_type', 'str', '0.1',
            'Specify N2 gas exchange scheme type'),
        ('N2O_exchange_present', 'bool', '1.1',
            'Is N2O gas exchange modeled ?'),
        ('N2O_exchange_type', 'str', '0.1',
            'Specify N2O gas exchange scheme type'),
        ('CFC11_exchange_present', 'bool', '1.1',
            'Is CFC11 gas exchange modeled ?'),
        ('CFC11_exchange_type', 'str', '0.1',
            'Specify CFC11 gas exchange scheme type'),
        ('CFC12_exchange_present', 'bool', '1.1',
            'Is CFC12 gas exchange modeled ?'),
        ('CFC12_exchange_type', 'str', '0.1',
            'Specify CFC12 gas exchange scheme type'),
        ('SF6_exchange_present', 'bool', '1.1',
            'Is SF6 gas exchange modeled ?'),
        ('SF6_exchange_type', 'str', '0.1',
            'Specify SF6 gas exchange scheme type'),
        ('13CO2_exchange_present', 'bool', '1.1',
            'Is 13CO2 gas exchange modeled ?'),
        ('13CO2_exchange_type', 'str', '0.1',
            'Specify 13CO2 gas exchange scheme type'),
        ('14CO2_exchange_present', 'bool', '1.1',
            'Is 14CO2 gas exchange modeled ?'),
        ('14CO2_exchange_type', 'str', '0.1',
            'Specify 14CO2 gas exchange scheme type'),
        ('other_gases', 'str', '0.1',
            'Specify any other gas exchange'),
        ]
    }

DETAILS['carbon_chemistry'] = {
    'description': 'Properties of carbon chemistry biogeochemistry',
    'properties': [
        ('type', 'ENUM:carbon_chemistry', '1.1',
            'Describe how carbon chemistry is modeled'),
        ('ph_scale', 'ENUM:ph_scale', '0.1',
            'If NOT OMIP protocol, describe pH scale.'),
        ('constants_if_not_OMIP', 'cs-str', '0.1',
            'If NOT OMIP protocol, list carbon chemistry constants.'),
        ]
    }

# --------------------------------------------------------------------
# SUB-PROCESS: TUNING APPLIED: Any tuning used to optimise the parameters
# --------------------------------------------------------------------
DETAILS['tuning_applied'] = {
    'description': 'Tuning methodology for ocean biogeochemistry component',
    'properties': [
        ('description', 'l-str', '1.1',
             "General overview description of tuning: explain and motivate the main targets and metrics retained. &"
             "Document the relative weight given to climate performance metrics versus process oriented metrics, &"
             "and on the possible conflicts with parameterization level tuning. In particular describe any struggle &"
             "with a parameter value that required pushing it to its limits to solve a particular model deficiency."),
        ('global_mean_metrics_used', 'cs-str', '0.1',
             "List set of metrics of the global mean state used in tuning model/component"),
        ('regional_metrics_used', 'cs-str', '0.1',
             "List of regional metrics of mean state used in tuning model/component"),
        ('trend_metrics_used', 'cs-str', '0.1',
             "List observed trend metrics used in tuning model/component"),
        ]
    }

# --------------------------------------------------------------------
# KEY PROPERTIES: ENUMERATIONS
# --------------------------------------------------------------------
ENUMERATIONS['model_family'] = {
    'description': 'Types of models for ocean biogeochemistry',
    'is_open': True,
    'members': [
        ('Geochemical', 'No living compartments'),
        ('NPZD', 'No plankton types'),
        ('PFT','Several plankton types'),
        ]
    }

ENUMERATIONS['elemental_stoichiometry_types'] = {
    'description': 'Types elemental stoichiometry for ocean biogeochemistry',
    'is_open': False,
    'members': [
        ('Fixed', 'Fixed stoichiometry'),
        ('Variable', 'Variable stoichiometry'),
        ('Mix of both','Both fixed and mixed stoichiometry'),
        ]
    }


ENUMERATIONS['passive_tracers_transport'] = {
    'description': 'Types of time stepping framework for passive tracers ocean biogeochemistry',
    'is_open': False,
    'members': [
        ('use ocean model transport time step', None),
        ('use specific time step', None),
        ]
    }

ENUMERATIONS['transport_types'] = {
    'description': 'Types of transport in ocean biogeochemistry',
    'is_open': False,
    'members': [
        ('Offline', None),
        ('Online', None),
        ]
    }

ENUMERATIONS['transport_scheme'] = {
    'description': 'Types of transport in ocean biogeochemistry',
    'is_open': True,
    'members': [
        ('Use that of ocean model', None),
        ]
    }

ENUMERATIONS['sources_atmos_deposition'] = {
    'description': 'Type of atmospheric deposition in ocean biogeochemistry',
    'is_open': False,
    'members': [
        ('from file (climatology)', None),
        ('from file (interannual variations)', None),
        ('from Atmospheric Chemistry model', None),
        ]
    }

ENUMERATIONS['sources_river_input'] = {
    'description': 'Type of river input in ocean biogeochemistry',
    'is_open': False,
    'members': [
        ('from file (climatology)', None),
        ('from file (interannual variations)', None),
        ('from Land Surface model', None),
        ]
    }

ENUMERATIONS['gas_exchange_types'] = {
    'description': 'Type of gas exchange in ocean biogeochemistry',
    'is_open': True,
    'members': [
        ('OMIP protocol', None),
        ]
    }

ENUMERATIONS['carbon_chemistry'] = {
    'description': 'Type of carbon chemistry in ocean biogeochemistry',
    'is_open': False,
    'members': [
        ('OMIP protocol', None),
        ('Other protocol', None),
        ]
    }

ENUMERATIONS['ph_scale'] = {
    'description': 'Type of carbon chemistry pH scale in ocean biogeochemistry',
    'is_open': True,
    'members': [
        ('Sea water', None),
        ('Free', None),
        ]
    }



"""
A realm process sepecialization.

For further information goto http://wordpress.es-doc.org/cmip6-model-specializations.
"""
# --------------------------------------------------------------------
# INTERNAL (do not change)
# --------------------------------------------------------------------
from collections import OrderedDict

DETAILS = OrderedDict()
ENUMERATIONS = OrderedDict()

# --------------------------------------------------------------------
# DESCRIPTION: Short description of the specialization.
# --------------------------------------------------------------------
DESCRIPTION = 'Sea Ice Radiative Processes'

# --------------------------------------------------------------------
# PROCESS: top level properties
# --------------------------------------------------------------------
DETAILS['toplevel'] = {
    'description': "Radiative processes in sea ice.",
    'properties': [
        ('surface_albedo', 'ENUM:seaice_albedo', '1.1',
            "Method used to handle surface albedo?"),
        ('ice_radiation_transmission', 'ENUM:ice_transmission', '1.N',
            "Method by which solar radiation through sea ice is handled?"),
        ]
    }

# --------------------------------------------------------------------
# PROCESS: ENUMERATIONS
# --------------------------------------------------------------------
ENUMERATIONS['seaice_albedo'] = {
    'description': "Surface albedo of sea ice component.",
    'is_open': True,
    'members': [
        ('Delta-Eddington', None),
        ('Parameterized', 'Sea ice albedo is parameterized.'),
        ('Multi-band albedo', 'Albedo value has a spectral dependence.'),
    ]
}

ENUMERATIONS['ice_transmission'] = {
    'description': "Ice radiative transmission.",
    'is_open': True,
    'members': [
        ('Delta-Eddington', None),
        ('Exponential attenuation', None),
        ('Ice radiation transmission per category', 'Radiation transmission through ice is different for each sea ice category.')
    ]
}



"""

A realm process specialization.

For further information goto http://wordpress.es-doc.org/cmip6-model-specializations.
"""
# --------------------------------------------------------------------
# INTERNAL (do not change)
# --------------------------------------------------------------------
from collections import OrderedDict

DETAILS = OrderedDict()
ENUMERATIONS = OrderedDict()

# --------------------------------------------------------------------
# DESCRIPTION: Short description of the specialization.
# --------------------------------------------------------------------
DESCRIPTION = 'Characteristics of the atmosphere radiation process'

# --------------------------------------------------------------------
# PROCESS: top level properties
# --------------------------------------------------------------------
DETAILS['toplevel'] = {
    'description': "Radiative agents in the atmosphere",
    'properties': [
        ('aerosols', 'ENUM:aerosol_types', '1.N',
            'Aerosols whose radiative effect is taken into account in the atmosphere model'),
        ]
    }

# --------------------------------------------------------------------
# SUB-PROCESS: shortwave_radiation
# --------------------------------------------------------------------
DETAILS['shortwave_radiation'] = {
    'description': 'Properties of the shortwave radiation scheme',
    'properties': [
        ('name', 'str', '0.1',
            'Commonly used name for the shortwave radiation scheme'),
        ('spectral_integration', 'ENUM:spectral_integration', '1.1',
            'Shortwave radiation scheme spectral integration'),
        ('transport_calculation', 'ENUM:transport_calculation', '1.N',
            'Shortwave radiation transport calculation methods'),
        ('spectral_intervals', 'int', '1.1',
            'Shortwave radiation scheme number of spectral intervals'),
        ('general_interactions', 'ENUM:radiative_interactions', '1.N',
            'General radiative interactions e.g. with aerosols, cloud ice and cloud water '),
        ]
    }

# --------------------------------------------------------------------
# SUB-PROCESS: shortwave_GHG
# --------------------------------------------------------------------
DETAILS['shortwave_GHG'] = {
    'description': 'Representation of greenhouse gases in the shortwave radiation scheme',
    'properties': [
        ('greenhouse_gas_complexity', 'ENUM:ghg_types', '1.N',
            'Complexity of greenhouse gases whose shortwave radiative effects are taken into account '
            'in the atmosphere model'),
        ('ODS', 'ENUM:ODS', '0.N',
            'Ozone depleting substances whose shortwave radiative effects are explicitly taken into account '
            'in the atmosphere model'),
        ('other_flourinated_gases', 'ENUM:other_fluorinated_gases', '0.N',
            'Other flourinated gases whose shortwave radiative effects are explicitly taken into account '
            'in the atmosphere model'),
        ]
}

# --------------------------------------------------------------------
# SUB-PROCESS: shortwave_cloud_ice
# --------------------------------------------------------------------
DETAILS['shortwave_cloud_ice'] = {
    'description': 'Shortwave radiative properties of ice crystals in clouds',
    'properties': [
        ('physical_representation', 'ENUM:cloud_ice_physical_representation', '1.N',
            'Physical representation of cloud ice crystals in the shortwave radiation scheme'),
        ('optical_methods', 'ENUM:optical_methods', '1.N',
            'Optical methods applicable to cloud ice crystals in the shortwave radiation scheme'),
    ]
}

# --------------------------------------------------------------------
# SUB-PROCESS: shortwave_cloud_liquid
# --------------------------------------------------------------------
DETAILS['shortwave_cloud_liquid'] = {
    'description': 'Shortwave radiative properties of liquid droplets in clouds',
    'properties': [
        ('physical_representation', 'ENUM:cloud_liquid_physical_representation', '1.N',
            'Physical representation of cloud liquid droplets in the shortwave radiation scheme'),
        ('optical_methods', 'ENUM:optical_methods_droplets', '1.N',
            'Optical methods applicable to cloud liquid droplets in the shortwave radiation scheme'),
    ]
}

# --------------------------------------------------------------------
# SUB-PROCESS: shortwave_cloud_inhomogeneity
# --------------------------------------------------------------------
DETAILS['shortwave_cloud_inhomogeneity'] = {
    'description': 'Cloud inhomogeneity in the shortwave radiation scheme',
    'properties': [
        ('cloud_inhomogeneity', 'ENUM:inhomogeneity_treatment', '1.1',
            'Method for taking into account horizontal cloud inhomogeneity'),
        ]
}

# --------------------------------------------------------------------
# SUB-PROCESS: shortwave_aerosols
# --------------------------------------------------------------------
DETAILS['shortwave_aerosols'] = {
    'description': 'Shortwave radiative properties of aerosols',
    'properties': [
        ('physical_representation', 'ENUM:aerosol_physical_representation', '1.N',
            'Physical representation of aerosols in the shortwave radiation scheme'),
        ('optical_methods', 'ENUM:optical_methods', '1.N',
            'Optical methods applicable to aerosols in the shortwave radiation scheme'),
    ]
}

# --------------------------------------------------------------------
# SUB-PROCESS: shortwave_gases
# --------------------------------------------------------------------
DETAILS['shortwave_gases'] = {
    'description': 'Shortwave radiative properties of gases',
    'properties': [
    ]
}

# --------------------------------------------------------------------
# SUB-PROCESS: longwave_radiation
# --------------------------------------------------------------------
DETAILS['longwave_radiation'] = {
    'description': 'Properties of the longwave radiation scheme',
    'properties': [
        ('name', 'str', '0.1',
            'Commonly used name for the longwave radiation scheme.'),
        ('spectral_integration', 'ENUM:spectral_integration', '1.1',
            'Longwave radiation scheme spectral integration'),
        ('transport_calculation', 'ENUM:transport_calculation', '1.N',
            'Longwave radiation transport calculation methods'),
        ('spectral_intervals', 'int', '1.1',
            'Longwave radiation scheme number of spectral intervals'),
        ('general_interactions', 'ENUM:radiative_interactions', '1.N',
            'General radiative interactions e.g. with aerosols, cloud ice and cloud water '),
        ]
    }

# --------------------------------------------------------------------
# SUB-PROCESS: longwave_GHG
# --------------------------------------------------------------------
DETAILS['longwave_GHG'] = {
    'description': 'Representation of greenhouse gases in the longwave radiation scheme',
    'properties': [
        ('greenhouse_gas_complexity', 'ENUM:ghg_types', '1.N',
            'Complexity of greenhouse gases whose longwave radiative effects are taken into account '
            'in the atmosphere model'),
        ('ODS', 'ENUM:ODS', '0.N',
            'Ozone depleting substances whose longwave radiative effects are explicitly taken into account '
            'in the atmosphere model'),
        ('other_flourinated_gases', 'ENUM:other_fluorinated_gases', '0.N',
            'Other flourinated gases whose longwave radiative effects are explicitly taken into account '
            'in the atmosphere model'),
        ]
}

# --------------------------------------------------------------------
# SUB-PROCESS: longwave_cloud_ice
# --------------------------------------------------------------------
DETAILS['longwave_cloud_ice'] = {
    'description': 'Longwave radiative properties of ice crystals in clouds',
    'properties': [
        ('physical_reprenstation', 'ENUM:cloud_ice_physical_representation', '1.N',
            'Physical representation of cloud ice crystals in the longwave radiation scheme'),
        ('optical_methods', 'ENUM:optical_methods', '1.N',
            'Optical methods applicable to cloud ice crystals in the longwave radiation scheme'),
    ]
}

# --------------------------------------------------------------------
# SUB-PROCESS: longwave_cloud_liquid
# --------------------------------------------------------------------
DETAILS['longwave_cloud_liquid'] = {
    'description': 'Longwave radiative properties of liquid droplets in clouds',
    'properties': [
        ('physical_representation', 'ENUM:cloud_liquid_physical_representation', '1.N',
            'Physical representation of cloud liquid droplets in the longwave radiation scheme'),
        ('optical_methods', 'ENUM:optical_methods_droplets', '1.N',
            'Optical methods applicable to cloud liquid droplets in the longwave radiation scheme'),
    ]
}

# --------------------------------------------------------------------
# SUB-PROCESS: longwave_cloud_inhomogeneity
# --------------------------------------------------------------------
DETAILS['longwave_cloud_inhomogeneity'] = {
    'description': 'Cloud inhomogeneity in the longwave radiation scheme',
    'properties': [
        ('cloud_inhomogeneity', 'ENUM:inhomogeneity_treatment', '1.1',
            'Method for taking into account horizontal cloud inhomogeneity'),
        ]
}

# --------------------------------------------------------------------
# SUB-PROCESS: longwave_aerosols
# --------------------------------------------------------------------
DETAILS['longwave_aerosols'] = {
    'description': 'Longwave radiative properties of aerosols',
    'properties': [
        ('physical_representation', 'ENUM:aerosol_physical_representation', '1.N',
            'Physical representation of aerosols in the longwave radiation scheme'),
        ('optical_methods', 'ENUM:optical_methods', '1.N',
            'Optical methods applicable to aerosols in the longwave radiation scheme'),
    ]
}

# --------------------------------------------------------------------
# SUB-PROCESS: longwave_gases
# --------------------------------------------------------------------
DETAILS['longwave_gases'] = {
    'description': 'Longwave radiative properties of gases',
    'properties': [
    ]
}

# --------------------------------------------------------------------
# PROCESS: ENUMERATIONS
# --------------------------------------------------------------------
ENUMERATIONS['aerosol_types'] = {
    'description': 'Aerosols whose radiative effect is taken into account in the atmospheric model.',
    'is_open': True,
    'members': [
        ('sulphate', None),
        ('nitrate', None),
        ('sea salt', None),
        ('dust', None),
        ('ice', None),
        ('organic', None),
        ('BC', 'black carbon / soot'),
        ('SOA', 'secondary organic aerosols'),
        ('POM', 'particulate organic matter'),
        ('polar stratospheric ice', None),
        ('NAT', 'nitric acid trihydrate'),
        ('NAD', 'nitric acid dihydrate'),
        ('STS', 'supercooled ternary solution aerosol particle'),
        ]
    }

ENUMERATIONS['ghg_types'] = {
    'description': 'Greenhouse gases whose radiative effect is taken into account in the atmosphere model',
    'is_open': True,
    'members': [
        ('CO2', 'Carbon Dioxide'),
        ('CH4', 'Methane'),
        ('N2O', 'Nitrous Oxide'),
        ('CFC-11 eq', 'Summarize the effect of non CO2, CH4, N2O and CFC-12 gases '
                      'with an equivalence concentration of CFC-11'),
        ('CFC-12 eq', 'Summarize the radiative effect of the Ozone Depleating Substances, ODSs, '
                      'with a CFC-12 equivalence concentration'),
        ('HFC-134a eq', 'Summarize the radiative effect of other fluorinated gases '
                         'with a HFC-134a equivalence concentration'),
        ('Explicit ODSs', 'Explicit representation of Ozone Depleting Substances '
                          'e.g. CFCs, HCFCs and Halons'),
        ('Explicit other fluorinated gases', 'Explicit representation of other fluorinated gases '
                                             'e.g. HFCs and PFCs'),
        ('O3', None),
        ('H2O', None),
        ]
    }

ENUMERATIONS['ODS'] = {
    'description': 'Ozone depleting substances, ODS, whose radiative effect is explicitly taken into account '
                   'in the atmosphere model',
    'is_open': True,
    'members': [
        ('CFC-12', 'CFC'),
        ('CFC-11', 'CFC'),
        ('CFC-113', 'CFC'),
        ('CFC-114', 'CFC'),
        ('CFC-115', 'CFC'),
        ('HCFC-22', 'HCFC'),
        ('HCFC-141b', 'HCFC'),
        ('HCFC-142b', 'HCFC'),
        ('Halon-1211', 'halon'),
        ('Halon-1301', 'halon'),
        ('Halon-2402', 'halon'),
        ('methyl chloroform', 'CH3CCl3'),
        ('carbon tetrachloride', 'CCl4'),
        ('methyl chloride', 'CH3Cl'),
        ('methylene chloride', 'CH2Cl2'),
        ('chloroform', 'CHCl3'),
        ('methyl bromide', 'Ch3Br'),
        ]
    }

ENUMERATIONS['other_fluorinated_gases'] = {
    'description': 'Other Fluorinated gases whose radiative effect is explicitly taken into account '
                   'in the atmosphere model',
    'is_open': True,
    'members': [
        ('HFC-134a', 'HFC'),
        ('HFC-23', 'HFC'),
        ('HFC-32', 'HFC'),
        ('HFC-125', 'HFC'),
        ('HFC-143a', 'HFC'),
        ('HFC-152a', 'HFC'),
        ('HFC-227ea', 'HFC'),
        ('HFC-236fa', 'HFC'),
        ('HFC-245fa', 'HFC'),
        ('HFC-365mfc', 'HFC'),
        ('HFC-43-10mee', 'HFC'),
        ('CF4', 'PFC'),
        ('C2F6', 'PFC'),
        ('C3F8', 'PFC'),
        ('C4F10', 'PFC'),
        ('C5F12', 'PFC'),
        ('C6F14', 'PFC'),
        ('C7F16', 'PFC'),
        ('C8F18', 'PFC'),
        ('c-C4F8', 'PFC'),
        ('NF3', None),
        ('SF6', None),
        ('SO2F2', None),
        ]
    }


ENUMERATIONS['spectral_integration'] = {
    'description': 'Spectral integration of the radiation scheme',
    'is_open': True,
    'members': [
        ('wide-band model', None),
        ('correlated-k', None),
        ('exponential sum fitting', None),
        ]
    }

ENUMERATIONS['transport_calculation'] = {
    'description': 'Radiation transport calculation methods',
    'is_open': True,
    'members': [
        ('two-stream', None),
        ('layer interaction', None),
        ('bulk', 'highly parameterised methods that use bulk expressions'),
        ('adaptive', 'exploits spatial and temporal correlations in optical characteristics'),
        ('multi-stream', None)
        ]
    }

ENUMERATIONS['radiative_interactions'] = {
    'description': 'General radiative interactions',
    'is_open': True,
    'members': [
        ('emission/absorption,', None),
        ('scattering', None),
    ]
}

ENUMERATIONS['cloud_ice_physical_representation'] = {
    'description': 'Physical representation of ice crystals in clouds',
    'is_open': True,
    'members': [
        ('bi-modal size distribution',
            'small mode diameters: a few tens of microns, large mode diameters: typically hundreds of microns'),
        ('ensemble of ice crystals', 'complex shapes represented with an ensemble of symmetric shapes'),
        ('mean projected area',
            'randomly oriented irregular ice crystals present a greater mean projected area than spheres'),
        ('ice water path', 'Integrated ice water path through the cloud kg m-2',),
        ('crystal asymmetry', None),
        ('crystal aspect ratio', None),
        ('effective crystal radius', None),
        ]
    }

ENUMERATIONS['cloud_liquid_physical_representation'] = {
    'description': 'Physical_radiative representation of liquid droplets in clouds',
    'is_open': True,
    'members': [
        ('cloud droplet number concentration', 'CDNC'),
        ('effective cloud droplet radii', None),
        ('droplet size distribution', None),
        ('liquid water path', 'Integrated liquid water path through the cloud kg m-2',),
        ]
    }

ENUMERATIONS['aerosol_physical_representation'] = {
    'description': 'Physical radiative representation of aerosols',
    'is_open': True,
    'members': [
        ('number concentration', None),
        ('effective radii', None),
        ('size distribution', None),
        ('asymmetry', None),
        ('aspect ratio', None),
        ('mixing state', 'For shortwave radiative interaction'),
        ]
    }

ENUMERATIONS['optical_methods'] = {
    'description': 'Optical methods used by radiation scheme',
    'is_open': True,
    'members': [
        ('T-matrix', 'For non-spherical particles'),
        ('geometric optics', 'For non-spherical particles'),
        ('finite difference time domain (FDTD)', 'For non-spherical particles'),
        ('Mie theory', 'For spherical particles'),
        ('anomalous diffraction approximation', None),
        ]
    }

ENUMERATIONS['optical_methods_droplets'] = {
    'description': 'Optical methods used by radiation scheme',
    'is_open': True,
    'members': [
        ('geometric optics', 'For non-spherical particles'),
        ('Mie theory', 'For spherical particles'),
        ]
}

ENUMERATIONS['cloud_ice_radiation_processes'] = {
    'description': 'Optical radiative processes for ice crystals in clouds',
    'is_open': True,
    'members': [
        ('emissivity', None),
        ('absorption', None),
        ('backward scattering', None),
        ('side scattering', None),
        ]
    }

ENUMERATIONS['cloud_liquid_radiation_processes'] = {
    'description': 'Optical radiative processes for liquid droplets in clouds',
    'is_open': True,
    'members': [
        ('droplet scattering', None),
        ('droplet absorption', None),
        ('broadband reflectivity', 'albedo'),
        ('broadband transmissivity', None),
        ('broadband absorbtivity', None),
        ]
    }

ENUMERATIONS['shortwave_scheme_type'] = {
    'description': 'Type of scheme used for shortwave radiation parameterisation',
    'is_open': True,
    'members': [
        ('wide-band model', None),
        ('bulk-scheme', 'highly parameterised methods that use bulk expressions'),
        ('two-stream', None),
        ('two-stream (delta-Eddington)', 'approximation for solar radiation calculations'),
        ]
    }

ENUMERATIONS['single_scattering_properties_methods'] = {
    'description': 'Methods for calculating single scattering properties of atmospheric constituents',
    'is_open': True,
    'members': [
        ('T-Matrix', 'an exact method'),
        ('geometrical optics', 'for particles that are much larger than the wavelength of light'),
        ('finite difference time domain', 'FDTD'),
        ('anomalous diffraction approximation', 'ADA'),
        ('k-distribution', None),
        ('band model', None),
        ('exponential sum fitting', None),
    ]
}


ENUMERATIONS['inhomogeneity_treatment'] = {
    'description': 'Cloud scheme inhomogeneity treatment',
    'is_open': True,
    'members': [
        ('Monte Carlo Independent Column Approximation', 'McICA'),
        ('Triplecloud', 'Regions of clear sky, optically thin cloud and optically thick cloud, Shonk et al 2010'),
        ('analytic', None),
        ]
    }




"""
.. module:: utils_factory.py
   :license: GPL/CeCIL
   :platform: Unix, Windows
   :synopsis: Takes specialization modules and returns instances of wrapper classes.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""
from utils import log
from utils_cache import set_specialization as cache_specialization
from utils_constants import *
from utils_model import EnumSpecialization
from utils_model import EnumChoiceSpecialization
from utils_model import PropertySetSpecialization
from utils_model import PropertySpecialization
from utils_model import ShortTable
from utils_model import ShortTableProperty
from utils_model import TopicSpecialization



def get_specialization(modules):
    """Returns a specialization wrapper.

    :param modules: 4 member tuple of python modules: root, grid, key-properties, processes.

    :returns: A specialization wrapper.
    :rtype: TopicSpecialization

    """
    # Create root topic.
    root = _create_topic(modules[0], None)

    # Create sub-topics.
    for module in modules[1:]:
        _create_topic(module, root)

    return root


create_specializations = get_specialization


def _create_topic(spec, parent, key=None):
    """Creates & returns a topic specialization.

    """
    if spec is None:
        return None

    # Hydrate either from a dictionary or a module.
    topic = TopicSpecialization(spec, parent)
    if isinstance(spec, dict):
        _set_topic_from_dict(topic, parent, key)
    else:
        _set_topic_from_module(topic, parent)

    # Set injected properties.
    _set_topic_injected_properties(topic)

    # Cache.
    cache_specialization(topic)

    return topic


def _set_topic_injected_properties(topic):
    """Injects a set of properties into the set of specializations.

    """
    # Model properties.
    if len(topic.path) == 3 and topic.path[1] == 'toplevel' and topic.path[2] == 'key_properties':
        if not topic.has_property('overview'):
            description = 'Top level overview of coupled model'
            _set_injected_property('overview', 'l-str', '1.1', description, topic)

        if not topic.has_property('keywords'):
            description = 'Keywords associated with coupled model'
            _set_injected_property('keywords', 'cs-str', '1.1', description, topic)

        if not topic.has_property('name'):
            description = 'Name of coupled model'
            _set_injected_property('name', 'str', '1.1', description, topic)

    # Topic key properties.
    elif len(topic.path) == 3 and topic.path[2] == 'key_properties':
        if not topic.has_property('overview'):
            description = 'Overview of {} model.'.format(topic.root.name)
            _set_injected_property('overview', 'l-str', '1.1', description, topic)

        if not topic.has_property('keywords'):
            description = 'Keywords associated with {} model code'.format(topic.root.name)
            _set_injected_property('keywords', 'cs-str', '1.1', description, topic)

        if not topic.has_property('name'):
            description = 'Name of {} model code'.format(topic.root.name)
            _set_injected_property('name', 'str', '1.1', description, topic)

    # Topic grid properties.
    elif len(topic.path) == 3 and topic.path[-1] == 'grid':
        if not topic.has_property('overview'):
            description = 'Overview of grid in {} model.'.format(topic.root.name)
            _set_injected_property('overview', 'l-str', '0.1', description, topic)

        if not topic.has_property('name'):
            description = 'Name of grid in {} model.'.format(topic.root.name)
            _set_injected_property('name', 'str', '0.1', description, topic)

    # Topic process properties.
    elif len(topic.path) == 3:
        if not topic.has_property('overview'):
            description = 'Overview of {} in {} model.'.format(topic.description.lower(), topic.root.name)
            _set_injected_property('overview', 'l-str', '0.1', description, topic)

        if not topic.has_property('name'):
            description = 'Commonly used name for the {} in {} model.'.format(topic.name_camel_case_spaced.lower(), topic.root.name)
            _set_injected_property('name', 'str', '0.1', description, topic)

    # Topic sub-process properties.
    elif len(topic.path) == 4:
        pass
        # if not topic.has_property('overview'):
        #     description = 'Overview of {} in {} model.'.format(topic.description.lower(), topic.root.name)
        #     _set_injected_property('overview', 'str', '0.1', description, topic)


def _set_injected_property(name, typeof, cardinality, description, topic):
    """Injects a property into the set of specializations.

    """
    try:
        enums = topic.spec.ENUMERATIONS
    except AttributeError:
        enums = topic.parent.spec.ENUMERATIONS
    _set_property(name, typeof, cardinality, description, topic, enums, False, True)


def _set_topic_from_module(topic, parent):
    """Set topic specialization attributes from a module.

    """
    try:
        topic.authors = topic.spec.AUTHORS
    except AttributeError:
        topic.authors = topic.root.authors
    try:
        topic.contact = topic.spec.CONTACT
    except AttributeError:
        topic.contact = topic.root.contact
    topic.description = topic.spec.DESCRIPTION
    if parent:
        topic.change_history = parent.change_history
        topic.contributors = parent.contributors
        topic.name = "_".join(topic.spec.__name__.split(".")[-1].split("_")[1:])
        topic.id = "{}.{}".format(parent.id, topic.name)
    else:
        topic.change_history = topic.spec.CHANGE_HISTORY
        topic.contributors = topic.spec.CONTRIBUTORS
        topic.name = topic.spec.__name__
        topic.id = "cmip6.{}".format(topic.name)

    # Assign properties / property sets.
    if hasattr(topic.spec, "DETAILS") and hasattr(topic.spec, "ENUMERATIONS"):
        for key, obj in topic.spec.DETAILS.items():
            # ... toplevel properties
            if key == "toplevel":
                _set_property_collection(topic, obj, topic.spec.ENUMERATIONS)

            # ... toplevel property sets
            elif key.startswith("toplevel"):
                _set_property_set(topic, key, obj, topic.spec.ENUMERATIONS)

            # ... sub-topic properties
            elif len(key.split(":")) == 1:
                # Create sub-sub-processes.
                _create_topic(obj, topic, key)
                _set_property_collection(topic.sub_topics[-1], obj, topic.spec.ENUMERATIONS)

            # ... sub-topic property sets
            elif len(key.split(":")) == 2:
                for st in topic.sub_topics:
                    if st.name == key.split(":")[0]:
                        _set_property_set(st, key, obj, topic.spec.ENUMERATIONS)


def _set_topic_from_dict(topic, parent, name):
    """Set topic specialization attributes from a dictionary.

    """
    topic.authors = parent.authors
    topic.contact = parent.contact
    topic.change_history = parent.change_history
    topic.contributors = parent.contributors
    topic.description = topic.spec['description']
    topic.id = "{}.{}".format(parent.id, name)
    topic.name = name


def _set_property_set(owner, key, obj, enumerations):
    """Set attributes of a property-set attributes from a dictionary.

    """
    ps = PropertySetSpecialization()
    ps.description = obj['description']
    ps.id = "{}.{}".format(owner.id, key.split(":")[-1])
    ps.key = key
    ps.name = key.split(":")[-1]
    ps.owner = owner
    _set_property_collection(ps, obj, enumerations)

    owner.property_sets.append(ps)

    # Cache.
    cache_specialization(ps)


def _set_property_collection(owner, obj, enumerations):
    """Set a collection of topic properties from a dictionary.

    """
    for name, typeof, cardinality, description in obj.get('properties', []):
        _set_property(name, typeof, cardinality, description, owner, enumerations, True, False)


def _set_property(name, typeof, cardinality, description, owner, enumerations, append, was_injected):
    """Returns a topic property.

    """
    p = PropertySpecialization()
    p.cardinality = cardinality
    p.description = description
    p.enum = _create_enum(p, typeof, enumerations) if typeof.startswith("ENUM:") else None
    p.id = "{}.{}".format(owner.id, name)
    p.was_injected = was_injected
    p.key = name
    p.name = name
    p.owner = owner
    p.typeof = typeof

    if append == True:
        owner.properties.append(p)
    else:
        owner.properties.insert(0, p)

    # Cache.
    cache_specialization(p)

    # Log injected.
    # if was_injected:
    #     log('injected property: {}'.format(p.id))

    return p


def _create_enum(detail, typeof, enumerations):
    """Creates & returns an enumeration specialzation wrapper.

    """
    key = typeof.split(":")[-1]
    obj = enumerations[key]

    e = EnumSpecialization()
    e.description = obj['description']
    e.detail = detail
    e.id = "{}.{}".format(detail.id, key)
    e.is_open = obj['is_open']
    e.label = key
    e.name = key
    e.id = key
    e.choices = [_create_enum_choice(e, i[0], i[1]) for i in obj.get('members', [])]

    return e


def _create_enum_choice(enum, value, description):
    """Creates & returns an enumeration choice specialzation wrapper.

    """
    ec = EnumChoiceSpecialization()
    ec.description = description
    ec.enum = enum
    ec.id = "{}.{}".format(enum.id, value)
    ec.value = value

    return ec


def get_short_tables(tables):
    """Returns a set of short table wrappers.

    :param tables: 2 member tuple: name, dict.

    :returns: A specialization wrapper.
    :rtype: tuple

    """
    return [_get_short_table(i, j) for i, j in tables]


create_short_tables = get_short_tables


def _get_short_table(name, obj):
    """Creates & returns a short-table wrapper.

    """
    result = ShortTable()
    result.authors = obj['AUTHORS']
    result.change_history = obj['CHANGE_HISTORY']
    result.contact = obj['CONTACT']
    result.contributors = obj['CONTRIBUTORS']
    result.label = obj['LABEL']
    result.name = name
    result.properties = [_get_short_table_property(i) for i in obj['PROPERTIES']]

    return result


def _get_short_table_property(obj):
    """Returns a short table property wrapper.

    :param modules: 4 member tuple of python modules: root, grid, key-properties, processes.

    :returns: A specialization wrapper.
    :rtype: tuple

    """
    result = ShortTableProperty()
    result.identifier = obj[0]
    result.priority = obj[1]

    return result



"""A grid specialization.

For further information goto http://wordpress.es-doc.org/cmip6-model-specializations.

"""
# --------------------------------------------------------------------
# INTERNAL (do not change)
# --------------------------------------------------------------------
from collections import OrderedDict

DETAILS = OrderedDict()
ENUMERATIONS = OrderedDict()

# --------------------------------------------------------------------
# DESCRIPTION: Short description of the specialization.
# --------------------------------------------------------------------
DESCRIPTION = 'Aerosol grid'

# --------------------------------------------------------------------
# PROCESS: top level
# --------------------------------------------------------------------
DETAILS['toplevel'] = {
    'description': 'Type of grid in the aerosol scheme',
    'properties': [
        ('matches_atmosphere_grid', 'bool', '1.1',
            ' Does the atmospheric aerosol grid match the atmosphere grid?'),
    ],
}

# --------------------------------------------------------------------
# SUB-PROCESS: RESOLUTION: The resolution of the grid.
# --------------------------------------------------------------------
DETAILS['resolution'] = {
    'description': 'Resolution in the atmospheric aerosol grid',
    'properties': [
        ('name', 'str', '1.1',
             "This is a string usually used by the modelling group to describe the resolution of this grid, e.g. ORCA025, N512L180, T512L70 etc."),
        ('canonical_horizontal_resolution', 'str', '0.1',
             "Expression quoted for gross comparisons of resolution, e.g. 50km or 0.1 degrees etc."),
        ('number_of_horizontal_gridpoints', 'int', '0.1',
             "Total number of horizontal (XY) points (or degrees of freedom) on computational grid."),
        ('number_of_vertical_levels', 'int', '0.1',
             "Number of vertical levels resolved on computational grid."),
        ('is_adaptive_grid', 'bool', '1.1',
             "Set to true if grid resolution changes during execution."),
    ],
}

# --------------------------------------------------------------------
# ENUMERATIONS
# --------------------------------------------------------------------



"""
.. module:: utils_loader.py
   :license: GPL/CeCIL
   :platform: Unix, Windows
   :synopsis: Loads CMIP6 specialization from file system.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""
import imp
import json
import os
import glob



def get_short_tables_definitions(input_dir, typeof):
    """Returns specialization short tables.

    :param str input_dir: Directory within which modules reside.
    :param str typeof: Type of specialization being processed.

    :returns: Tuple of decoded short tables.
    :rtype: tuple

    """
    def _decode(fpath):
        name = fpath.split("/")[-1].split(".")[0].replace("_short_table", "")
        with open(fpath, 'r') as fstream:
            return name, json.loads(fstream.read())

    fpaths = os.path.join(input_dir, "short_tables")
    fpaths = os.path.join(fpaths, "{}_*.json".format(typeof))
    tables = [_decode(i) for i in glob.glob(fpaths)]

    return sorted(tables, key=lambda i: i[0])


def get_modules(input_dir, typeof):
    """Returns specialization modules.

    :param str input_dir: Directory within which modules reside.
    :param str typeof: Type of specialization being processed.

    """
    # Load specialization modules.
    modules = _get_modules(input_dir, typeof)
    if not modules:
        raise KeyError("Specializations not found")

    # Set specializations.
    root = _get_module(modules, typeof)
    children = [_get_module(modules, i) for i in [
        '{}_key_properties'.format(typeof),
        '{}_grid'.format(typeof)
    ] + root.PROCESSES]

    return [root] + [i for i in children if i is not None]


def _get_modules(input_dir, typeof):
    """Returns a set of specialization modules.

    """
    def _is_target(fname):
        return not fname.startswith('_') and \
               fname.endswith('.py') and \
               fname.startswith(typeof)

    modules = sorted([i for i in os.listdir(input_dir) if _is_target(i)])
    modules = [os.path.join(input_dir, m) for m in modules]
    modules = [(m.split("/")[-1].split(".")[0], m) for m in modules]
    modules = [imp.load_source(i, j) for i, j in modules]

    return modules


def _get_module(modules, name):
    """Returns a specialization module.

    """
    for module in modules:
        if module.__name__ == name:
            return module



"""A key-properties specialization.

For further information goto http://wordpress.es-doc.org/cmip6-model-specializations.

"""
# --------------------------------------------------------------------
# INTERNAL (do not change)
# --------------------------------------------------------------------
from collections import OrderedDict

DETAILS = OrderedDict()
ENUMERATIONS = OrderedDict()

# --------------------------------------------------------------------
# DESCRIPTION: Short description of the specialization.
# --------------------------------------------------------------------
DESCRIPTION = 'Ocean key properties'

# --------------------------------------------------------------------
# KEY PROPERTIES: top level
# --------------------------------------------------------------------
DETAILS['toplevel'] = {
    'description': 'General key properties in ocean',
    'properties': [
        ('model_family', 'ENUM:model_family_types', '1.1',
            'Type of ocean model.'),
        ('basic_approximations', 'ENUM:ocean_basic_approx_types', '1.N',
            'Basic approximations made in the ocean.',),
        ('prognostic_variables', 'ENUM:prognostic_vars_types', '1.N',
            'List of prognostic variables in the ocean component.'),
        ]
    }

DETAILS['toplevel:seawater_properties'] = {
    'description': 'Physical properties of seawater in ocean',
    'properties' : [
        ('eos_type', 'ENUM:seawater_eos_types', '1.1',
            'Type of EOS for sea water'),
        ('eos_functional_temp', 'ENUM:seawater_eos_func_temp', '1.1',
            'Temperature used in EOS for sea water'),
        ('eos_functional_salt', 'ENUM:seawater_eos_func_salt', '1.1',
            'Salinity used in EOS for sea water'),
        ('eos_functional_depth', 'ENUM:seawater_eos_func_depth', '1.1',
            'Depth or pressure used in EOS for sea water ?'),
        ('ocean_freezing_point', 'ENUM:seawater_freezing_point', '1.1',
            'Equation used to compute the freezing point (in deg C) of seawater, as a function of salinity and pressure'),
        ('ocean_specific_heat', 'float', '1.1',
            'Specific heat in ocean (cpocean) in J/(kg K)'),
        ('ocean_reference_density', 'float', '1.1',
            'Boussinesq reference density (rhozero) in kg / m3'),
        ]
    }

DETAILS['toplevel:bathymetry'] = {
    'description': 'Properties of bathymetry in ocean',
    'properties' : [
        ('reference_dates', 'ENUM:bathymetry_ref_dates', '1.1',
            'Reference date of bathymetry'),
        ('type', 'bool', '1.1',
            'Is the bathymetry fixed in time in the ocean ?'),
        ('ocean_smoothing', 'l-str', '1.1',
            'Describe any smoothing or hand editing of bathymetry in ocean'),
        ('source', 'str', '1.1',
            'Describe source of bathymetry in ocean'),
        ]
    }

DETAILS['toplevel:nonoceanic_waters'] = {
    'description': 'Non oceanic waters treatement in ocean',
    'properties': [
        ('isolated_seas','l-str', '0.1',
            'Describe if/how isolated seas is performed'),
        ('river_mouth','l-str', '0.1',
            'Describe if/how river mouth mixing or estuaries specific treatment is performed'),
       ]
    }

DETAILS['toplevel:software_properties'] = {
    'description': 'Software properties of ocean code',
    'properties':[
        ('repository','str', '0.1',
            "Location of code for this component."),
        ('code_version','str', '0.1',
            "Code version identifier."),
        ('code_languages','cs-str', '0.1',
            "Code language(s)."),
    ]
}

# --------------------------------------------------------------------
# SUB-PROCESS: RESOLUTION: The resolution of the grid.
# --------------------------------------------------------------------
DETAILS['resolution'] = {
    'description': 'Resolution in the ocean grid',
    'properties': [
        ('name', 'str', '1.1',
             "This is a string usually used by the modelling group to describe the resolution of this grid, e.g. ORCA025, N512L180, T512L70 etc."),
        ('canonical_horizontal_resolution', 'str', '1.1',
             "Expression quoted for gross comparisons of resolution, eg. 50km or 0.1 degrees etc."),
        ('range_horizontal_resolution', 'str', '1.1',
             "Range of horizontal resolution with spatial details, eg. 50(Equator)-100km or 0.1-0.5 degrees etc."),
        ('number_of_horizontal_gridpoints', 'int', '1.1',
             "Total number of horizontal (XY) points (or degrees of freedom) on computational grid."),
        ('number_of_vertical_levels', 'int', '1.1',
             "Number of vertical levels resolved on computational grid."),
        ('is_adaptive_grid', 'bool', '1.1',
             "Default is False. Set true if grid resolution changes during execution."),
        ('thickness_level_1','float', '1.1',
             "Thickness of first surface ocean level (in meters)")
        ],
    }

# --------------------------------------------------------------------
# SUB-PROCESS: TUNING APPLIED: Any tuning used to optimise the parameters
# --------------------------------------------------------------------
DETAILS['tuning_applied'] = {
    'description': 'Tuning methodology for ocean component',
    'properties': [
        ('description', 'l-str', '1.1',
             """General overview description of tuning: explain and motivate the main targets and metrics retained.
             Document the relative weight given to climate performance metrics versus process oriented metrics,
             and on the possible conflicts with parameterization level tuning. In particular describe any struggle
             with a parameter value that required pushing it to its limits to solve a particular model deficiency.
             """),
        ('global_mean_metrics_used', 'cs-str', '0.1',
             "List set of metrics of the global mean state used in tuning model/component"),
        ('regional_metrics_used', 'cs-str', '0.1',
             "List of regional metrics of mean state (e.g THC, AABW, regional means etc) used in tuning model/component"),
        ('trend_metrics_used', 'cs-str', '0.1',
             "List observed trend metrics used in tuning model/component"),
        ]
    }

# --------------------------------------------------------------------
# SUB-PROCESS: CONSERVATION PROPERTIES: Details of methodology needed to conserve variables between processes
# --------------------------------------------------------------------
DETAILS['conservation'] = {
    'description': 'Conservation in the ocean component',
    'properties': [
        ('description', 'l-str', '1.1',
             'Brief description of conservation methodology'),
        ('scheme', 'ENUM:conservation_props_types', '1.N',
            'Properties conserved in the ocean by the numerical schemes'),
        ('consistency_properties', 'cs-str','0.1',
            'Any additional consistency properties (energy conversion, pressure gradient discretisation, ...)?'),
        ('corrected_conserved_prognostic_variables', 'cs-str', '0.1', # Can we constrains these variable
             "Set of variables which are conserved by *more* than the numerical scheme alone."),
        ('was_flux_correction_used', 'bool', '0.1',
             "Does conservation involve flux correction ?"),
        ],
    }

# --------------------------------------------------------------------
# KEY PROPERTIES: ENUMERATIONS
# --------------------------------------------------------------------
ENUMERATIONS['model_family_types'] = {
    'description': 'Types of ocean models',
    'is_open': True,
    'members': [
        ('OGCM', None),
        ('slab ocean', None),
        ('mixed layer ocean', None)
        ]
    }

ENUMERATIONS['ocean_basic_approx_types'] = {
    'description': 'Types of basic approximation in ocean',
    'is_open': True,
    'members': [
        ('Primitive equations', None),
        ('Non-hydrostatic', None),
        ('Boussinesq', None)
        ]
    }

ENUMERATIONS['prognostic_vars_types'] = {
    'description': 'List of prognostic variables in ocean',
    'is_open': True,
    'members': [
        ('Potential temperature', None),
        ('Conservative temperature', None),
        ('Salinity', None),
        ('U-velocity', None),
        ('V-velocity', None),
        ('W-velocity', None),
        ('SSH', 'Sea Surface Height')
        ]
    }

ENUMERATIONS['seawater_eos_types'] = {
    'description': 'Types of seawater Equation of State in ocean',
    'is_open': True,
    'members': [
        ('Linear', None),
        ('Wright, 1997', None),
        ('Mc Dougall et al.', None),
        ('Jackett et al. 2006', None),
        ('TEOS 2010', None)
        ]
    }

ENUMERATIONS['seawater_eos_func_temp'] = {
    'description': 'Types of temperature used in EOS in ocean',
    'is_open': False,
    'members': [
        ('Potential temperature', None),
        ('Conservative temperature', None),
        ]
    }

ENUMERATIONS['seawater_eos_func_salt'] = {
    'description': 'Types of salinity used in EOS in ocean',
    'is_open': False,
    'members': [
        ('Practical salinity Sp', None),
        ('Absolute salinity Sa', None),
        ]
    }

ENUMERATIONS['seawater_eos_func_depth'] = {
    'description': 'Types of depth used in EOS in ocean',
    'is_open': False,
    'members': [
        ('Pressure (dbars)', None),
        ('Depth (meters)', None),
        ]
    }

ENUMERATIONS['seawater_freezing_point'] = {
    'description': 'Types of seawater freezing point equation in ocean',
    'is_open': True,
    'members': [
        ('TEOS 2010', None)
        ]
    }

ENUMERATIONS['bathymetry_ref_dates'] = {
    'description': 'List of reference dates for bathymetry in ocean',
    'is_open': True,
    'members': [
        ('Present day', None),
        ('21000 years BP', None),
        ('6000 years BP', None),
        ('LGM', 'Last Glacial Maximum'),
        ('Pliocene', None)
        ]
    }

ENUMERATIONS['conservation_props_types'] = {
    'description': 'List of properties that can be conserved in ocean',
    'is_open': True,
    'members': [
        ('Energy', None),
        ('Enstrophy', None),
        ('Salt', None),
        ('Volume of ocean', None),
        ('Momentum', None)
        ]
    }



"""A process specialization.

For further information goto http://wordpress.es-doc.org/cmip6-model-specializations.

"""
# --------------------------------------------------------------------
# INTERNAL (do not change)
# --------------------------------------------------------------------
from collections import OrderedDict

DETAILS = OrderedDict()
ENUMERATIONS = OrderedDict()

# --------------------------------------------------------------------
# DESCRIPTION: Short description of the specialization.
# --------------------------------------------------------------------
DESCRIPTION = 'Aerosol model'

# --------------------------------------------------------------------
# PROCESS: top level
# --------------------------------------------------------------------
DETAILS['toplevel'] = {
    'description': 'Top level aerosol model properties',
    'properties': [
        ('overview', 'l-str', '1.1',
            'Overview of atmospheric aerosol model'),
        ('processes', 'ENUM:model_processes', '1.N',
            'Processes included in the aerosol model.'),
        ('coupling', 'ENUM:coupling', '0.N',
            'Other model components coupled to the aerosol model'),
        ('gas_phase_precursors', 'ENUM:gas_phase_precursors', '1.N',
            'Gas phase aerosol precursors.'),
        ('scheme_type', 'ENUM:scheme_types', '1.N',
             'Type(s) of aerosol scheme used by the aerosol model (potentially multiple: some species may be covered by one type of aerosol scheme and other species covered by another type).'),
        ('bulk_scheme_species', 'ENUM:species', '1.N',
             'Species covered by the bulk scheme.'),
    ]
}

# --------------------------------------------------------------------
# ENUMERATIONS
# --------------------------------------------------------------------
ENUMERATIONS['model_processes'] = {
    'description': 'Processes included in the aerosol model.',
    'is_open': False,
    'members': [
        ('Dry deposition', None),
        ('Sedimentation', None),
        ('Wet deposition (impaction scavenging)', None),
        ('Wet deposition (nucleation scavenging)', None),
        ('Coagulation', None),
        ('Oxidation (gas phase)', None),
        ('Oxidation (in cloud)', None),
        ('Condensation', None),
        ('Ageing', None),
        ('Advection (horizontal)', None),
        ('Advection (vertical)', None),
        ('Heterogeneous chemistry', None),
        ('Nucleation', None),
    ]
}

ENUMERATIONS['coupling'] = {
    'description':'Other model components coupled to the aerosol model.',
    'is_open': True,
    'members':[
        ('Radiation', None),
        ('Land surface', None),
        ('Heterogeneous chemistry', None),
        ('Clouds', None),
        ('Ocean', None),
        ('Cryosphere', None),
        ('Gas phase chemistry', None),
    ]
}

ENUMERATIONS['gas_phase_precursors'] = {
    'description': 'Gas phase aerosol precursors.',
    'is_open': True,
    'members':[
        ('DMS', None),
        ('SO2', None),
        ('Ammonia', None),
        ('Iodine', None),
        ('Terpene', None),
        ('Isoprene', None),
        ('VOC', None),
        ('NOx', None),
    ]
}

ENUMERATIONS['scheme_types'] = {
    'description': 'Type(s) of aerosol scheme used by the aerosols model (potentially multiple: some species may be covered by one type of aerosol scheme and other species covered by another type).',
    'is_open': True,
    'members':[
        ('Bulk', None),
        ('Modal', None),
        ('Bin', None),
    ]
}

ENUMERATIONS['species'] = {
    'description':'Aerosol species.',
    'is_open': True,
    'members':[
        ('Sulphate', None),
        ('Nitrate', None),
        ('Sea salt', None),
        ('Dust', None),
        ('Ice', None),
        ('Organic', None),
        ('Black carbon / soot', None),
        ('SOA (secondary organic aerosols)', None),
        ('POM (particulate organic matter)', None),
        ('Polar stratospheric ice', None),
        ('NAT (Nitric acid trihydrate)', None),
        ('NAD (Nitric acid dihydrate)', None),
        ('STS (supercooled ternary solution aerosol particule)', None),
    ]
}




"""A process specialization.

For further information goto http://wordpress.es-doc.org/cmip6-model-specializations.

"""
# --------------------------------------------------------------------
# INTERNAL (do not change)
# --------------------------------------------------------------------
from collections import OrderedDict

DETAILS = OrderedDict()
ENUMERATIONS = OrderedDict()

# --------------------------------------------------------------------
# DESCRIPTION: Short description of the specialization.
# --------------------------------------------------------------------
DESCRIPTION = 'Ocean advection'

# --------------------------------------------------------------------
# SUB-PROCESS: Momentum.
# --------------------------------------------------------------------
DETAILS['momentum'] = {
    'description': 'Properties of lateral momemtum advection scheme in ocean',
    'properties': [
        ('type', 'ENUM:adv_mom_scheme_types', '1.1',
            'Type of lateral momemtum advection scheme in ocean'),
        ('scheme_name', 'str', '1.1',
            'Name of ocean momemtum advection scheme'),
        ('ALE', 'bool', '0.1',
            'Using ALE for vertical advection ? (if vertical coordinates are sigma)'),
        ]
    }

# --------------------------------------------------------------------
# SUB-PROCESS: Lateral tracers.
# --------------------------------------------------------------------
DETAILS['lateral_tracers'] = {
    'description':'Properties of lateral tracer advection scheme in ocean',
    'properties': [
        ('order', 'int', '1.1',
            "Order of lateral tracer advection scheme in ocean"),
        ('flux_limiter', 'bool', '1.1',
            "Monotonic flux limiter for lateral tracer advection scheme in ocean ?"),
        ('effective_order', 'float', '1.1',
            "Effective order of limited lateral tracer advection scheme in ocean"),
        ('name', 'str', '1.1',
            "Descriptive text for lateral tracer advection scheme in ocean (e.g. MUSCL, PPM-H5, PRATHER,...)"),
        ('passive_tracers','ENUM:passive_tracers_list','0.N',
            "Passive tracers advected"),
        ('passive_tracers_advection','str','0.1',
            "Is advection of passive tracers different than active ? if so, describe.")
        ]
    }

# --------------------------------------------------------------------
# SUB-PROCESS: Vertical tracers.
# --------------------------------------------------------------------
DETAILS['vertical_tracers'] = {
    'description': 'Properties of vertical tracer advection scheme in ocean',
    'properties': [
        ('name', 'str', '1.1',
            "Descriptive text for vertical tracer advection scheme in ocean (e.g. MUSCL, PPM-H5, PRATHER,...)"),
        ('flux_limiter', 'bool', '1.1',
            "Monotonic flux limiter for vertical tracer advection scheme in ocean ?"),
        ]
    }

# --------------------------------------------------------------------
# ENUMERATIONS
# --------------------------------------------------------------------
ENUMERATIONS['adv_mom_scheme_types'] = {
    'description': 'Type of lateral momemtum advection scheme in ocean',
    'is_open': False,
    'members': [
        ('Flux form', None),
        ('Vector form', None)
        ]
    }

ENUMERATIONS['passive_tracers_list'] = {
    'description': 'Passive tracers in ocean',
    'is_open': True,
    'members': [
        ('Ideal age', None),
        ('CFC 11', None),
        ('CFC 12', None),
        ('SF6', None)
        ]
    }



"""A realm key-properties specialization.

For further information goto http://wordpress.es-doc.org/cmip6-model-specializations.

"""
# --------------------------------------------------------------------
# INTERNAL (do not change)
# --------------------------------------------------------------------
from collections import OrderedDict

DETAILS = OrderedDict()
ENUMERATIONS = OrderedDict()

# --------------------------------------------------------------------
# DESCRIPTION: Short description of the specialization.
# --------------------------------------------------------------------
DESCRIPTION = 'Key properties of the aerosol model'

# --------------------------------------------------------------------
# PROCESS: top level
# --------------------------------------------------------------------
DETAILS['toplevel'] = {
    'description': 'Top level key properties in aerosol model',
    'properties': [
        ('scheme_scope', 'ENUM:scheme_scopes', '1.N',
            'Atmospheric domains covered by the aerosol model'),
        ('basic_approximations', 'str', '1.1',
            'Basic approximations made in the aerosol model',),
        ('prognostic_variables_form', 'ENUM:prognostic_vars_types', '1.N',
            'Prognostic variables in the aerosol model'),
        ('number_of_tracers', 'int', '1.1',
            'Number of tracers in the aerosol model'),
        ('family_approach', 'bool', '1.1',
            'Are aerosol calculations generalized into families of species?'),
    ]
}

# --------------------------------------------------------------------
# SUB-PROCESS: timestepping framework
# --------------------------------------------------------------------
DETAILS['timestep_framework'] = {
    'description': 'Physical properties of seawater in ocean',
    'properties' : [
        ('method', 'ENUM:timestepping_methods', '1.1',
            'Mathematical method deployed to solve the time evolution of the prognostic variables'),
        ('split_operator_advection_timestep', 'int', '0.1',
            'Timestep for aerosol advection (in seconds)'),
        ('split_operator_physical_timestep', 'int', '0.1',
            'Timestep for aerosol physics (in seconds).'),
        ('integrated_timestep', 'int', '1.1',
            'Timestep for the aerosol model (in seconds)'),
        ('integrated_scheme_type', 'ENUM:integrated_scheme_types', '1.1',
            'Specify the type of timestep scheme'),
    ],
}

# --------------------------------------------------------------------
# SUB-PROCESS: Meteorological forcings
# --------------------------------------------------------------------
DETAILS['meteorological_forcings'] = {
    'description': '',
    'properties' : [
        ('variables_3D', 'cs-str', '0.1',
            'Three dimensional forcing variables, e.g. U, V, W, T, Q, P, conventive mass flux'),
        ('variables_2D', 'cs-str', '0.1',
            'Two dimensional forcing variables, e.g. land-sea mask definition'),
        ('frequency', 'int', '0.1',
            'Frequency with which meteorological forcings are applied (in seconds).'),
        ]
}

# --------------------------------------------------------------------
# SUB-PROCESS: RESOLUTION: The resolution of the grid.
# --------------------------------------------------------------------
DETAILS['resolution'] = {
    'description': 'Resolution in the aerosol model grid',
    'properties': [
        ('name', 'str', '1.1',
             "This is a string usually used by the modelling group to describe the resolution of this grid, e.g. ORCA025, N512L180, T512L70 etc."),
        ('canonical_horizontal_resolution', 'str', '0.1',
             "Expression quoted for gross comparisons of resolution, eg. 50km or 0.1 degrees etc."),
        ('number_of_horizontal_gridpoints', 'int', '0.1',
             "Total number of horizontal (XY) points (or degrees of freedom) on computational grid."),
        ('number_of_vertical_levels', 'int', '0.1',
             "Number of vertical levels resolved on computational grid."),
        ('is_adaptive_grid', 'bool', '1.1',
             "Set to true if the grid resolution changes during execution."),
    ],
}

# --------------------------------------------------------------------
# SUB-PROCESS: TUNING APPLIED: Any tuning used to optimise the parameters
# --------------------------------------------------------------------
DETAILS['tuning_applied'] = {
    'description': 'Tuning methodology for aerosol model',
    'properties': [
        ('description', 'l-str', '1.1',
             "General overview description of tuning: explain and motivate the main targets and metrics retained. Document the relative weight given to climate performance metrics versus process oriented metrics, and on the possible conflicts with parameterization level tuning. In particular describe any struggle with a parameter value that required pushing it to its limits to solve a particular model deficiency."),
        ('global_mean_metrics_used', 'cs-str', '0.1',
             "List of metrics of the global mean state used in tuning model/component"),
        ('regional_metrics_used', 'cs-str', '0.1',
             "List of metrics of regional mean state used in tuning model/component"),
        ('trend_metrics_used', 'cs-str', '0.1',
             "List observed trend metrics used in tuning model/component"),
    ]
}

DETAILS['toplevel:software_properties'] = {
    'description': 'Software properties of aerosol code',
    'properties':[
        ('repository','str', '0.1',
            "Location of code for this component."),
        ('code_version','str', '0.1',
            "Code version identifier."),
        ('code_languages','cs-str', '0.1',
            "Code language(s)."),
    ]
}

# --------------------------------------------------------------------
# KEY PROPERTIES: ENUMERATIONS
# --------------------------------------------------------------------
ENUMERATIONS['scheme_scopes'] = {
    'description': 'Atmospheric domains covered by the aerosol model',
    'is_open': True,
    'members': [
        ('troposphere', None),
        ('stratosphere', None),
        ('mesosphere', None),
        ('whole atmosphere', None),
    ]
}

ENUMERATIONS['prognostic_vars_types'] = {
    'description': 'Form of prognostic variables in the aerosol model.',
    'is_open': True,
    'members': [
        ('3D mass/volume ratio for aerosols', None),
        ('3D number concentration for aerosols', None),
    ]
}

ENUMERATIONS['timestepping_methods'] = {
    'description': 'Mathematical method deployed to solve the time evolution of the prognostic variables.',
    'is_open': True,
    'members': [
        ('Uses atmospheric chemistry time stepping', None),
        ('Specific timestepping (operator splitting)', None),
        ('Specific timestepping (integrated)', None),
    ]
}

ENUMERATIONS['integrated_scheme_types'] = {
    'description': 'Specify the type of timestep scheme',
    'is_open': True,
    'members': [
        ('Explicit', None),
        ('Implicit', None),
        ('Semi-implicit', None),
        ('Semi-analytic', None),
        ('Impact solver', None),
        ('Back Euler', None),
        ('Newton Raphson', None),
        ('Rosenbrock', None),
    ]
}
